
import sys, argparse, glob, collections
from functools import partial

#import os
#os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "TRUE"
#os.environ["QT_SCALE_FACTOR"] = "2.5"

import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore

# needed right after QT imports for high-DPI screens
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

from .window import Window, set_style
from .experiments import Experiment, check_new_experiments
from .plots import Plots
from .visualizations import Visualizations


def main():
  # parse command line options
  parser = argparse.ArgumentParser()
  parser.add_argument("folder", help="Root folder where experiments are found.")
  parser.add_argument("-plotsize", default=0, type=float, help="Initial size of plots, in pixels.")
  #parser.add_argument("-smoothen", default=0, type=float)
  parser.add_argument("-mpl-dpi", default=100, type=int, help="DPI setting for MatPlotLib plots, may be used if text is too big/small (useful for high-DPI monitors).")
  parser.add_argument("--force-reopen-files", action='store_true', default=False, help="Slower but more reliable refresh method, useful for remote files.")
  parser.add_argument("-refresh-plots", default=1999, type=int, help="Refresh interval for plot updates, in miliseconds.")
  parser.add_argument("-refresh-new", default=3999, type=int, help="Refresh interval for finding new experiments, in miliseconds.")
  parser.add_argument("-refresh-vis", default=2999, type=int, help="Refresh interval for visualizations, in miliseconds.")
  parser.add_argument("--no-vis-snapshot", action='store_true', default=False, help="Visualizations are draw using a snapshot of the visualization function, saved with each experiment. This ensures visualizations from old experiments are maintained. Passing this option disables this behavior, which may be useful for debugging.")
  args = parser.parse_args()
  
  # find experiment files
  files = glob.glob(args.folder + "/**/stats.csv", recursive=True)

  # load the experiments
  experiments = [Experiment(filename, args.folder, args.force_reopen_files) for filename in files]

  # create Qt application
  app = QtWidgets.QApplication(sys.argv)
  set_style(app)

  # create window and plots holders
  window = Window(args)
  plots = Plots(window)
  visualizations = Visualizations(window, args.no_vis_snapshot, args.mpl_dpi)

  # create initial plots for all the experiments
  for exp in experiments:
    window.add_experiment(exp, refresh_table=False)
  
  if len(experiments) > 0:  # select first row if any
    window.table.selectRow(0)
  
  # create timer for updating the plots periodically if needed
  plot_timer = QtCore.QTimer()
  plot_timer.timeout.connect(partial(plots.update_plots, experiments))
  plot_timer.start(args.refresh_plots)
  
  # create timer to check for new experiments
  new_exp_timer = QtCore.QTimer()
  new_exp_timer.timeout.connect(partial(check_new_experiments, experiments, set(files), args.folder, window, args.force_reopen_files))
  new_exp_timer.start(args.refresh_new)
  
  # create timer for updating the current visualizations
  vis_timer = QtCore.QTimer()
  vis_timer.timeout.connect(visualizations.update)
  vis_timer.start(args.refresh_vis)

  window.show()
  
  window.sort_table_by_timestamp(refresh_table=True)

  app.exec_()
  #sys.exit(app.exec_())


if __name__ == "__main__":
  main()




import collections, json, glob, re, logging
import numpy as np
from os import path as os_path

class Experiment():
  def __init__(self, filename, base_folder, force_reopen_files):
    directory = os_path.dirname(filename)
    self.name = os_path.relpath(directory, base_folder)  # experiment name is the path relative to base folder
    self.filename = filename
    self.directory = directory

    # read JSON file for metadata (including timestamp), if it exists
    try:
      with open(directory + '/meta.json', 'r') as file:
        self.meta = json.load(file)
    except IOError:
      self.meta = {}

    self.names = []
    self.data = []
    
    # store whether the experiment has finished or not
    self.done = self.meta.get('_done', False)

    self.read_data = None
    self.visible = True
    self.is_selected = False
    self.style = {}  # used by Plots
    self.style_order = 1000

    # do first read
    self.read_data = self.read_data_slow() if force_reopen_files else self.read_data_fast()
    try:
      next(self.read_data)
    except StopIteration:
      pass
    except IOError:
      logging.exception('Error reading ' + self.filename)
  
  def read_data_fast(self):  # generator function
    # polls the file for new data, bu keeps the file handle open
    with open(self.filename, 'r') as file:
      has_new_data = False

      while True:
        # try to read new line. need to restore read position when there's no new
        # line to read, to cope with incomplete lines (while they're being written)
        where = file.tell()
        if self.read_single_line(file):
          has_new_data = True  # keep reading
        else:
          if self.done: return
          yield has_new_data

          has_new_data = False
          file.seek(where)  # back to beginning of line

  def read_data_slow(self):  # generator function
    # initial read
    with open(self.filename, 'r') as file:
      while self.read_single_line(file):
        pass
      where = file.tell()  # position for next read
    
    # detect when file size changes (note file handle is closed most of the time)
    old_size = 0
    while not self.done:
      new_size = os_path.getsize(self.filename)
      file_changed = (new_size != old_size)
      old_size = new_size

      if file_changed:
        # get new data from the file, picking up where we left off. don't use file size, depends on OS.
        with open(self.filename, 'r') as file:
          file.seek(where)
          while self.read_single_line(file):
            pass
          where = file.tell()  # remember where we left off for next time

      yield file_changed

  def read_single_line(self, file):
    # reads a single CSV line from the file, returning True if some data was read.
    line = file.readline()
    if line:
      if len(self.names) == 0:
        # read CSV file header (with stat names). they're separated by commas, which can be escaped: \,
        self.names = re.split(r'(?<!\\),', line.strip())

        if len(self.names) < 2: raise IOError('CSV has too few headers.')

        self.data = [[] for _ in range(len(self.names))]
      else:
        # read line of stat values
        values = line.split(',')

        if len(values) == len(self.names):
          for i in range(len(values)):  # read one value per column
            self.data[i].append(float(values[i]))
        else:
          if len(values) > len(self.names):
            raise IOError('CSV line has more cells than the header.')
          return False  # incomplete line, file may still be written to
        
      return True
    return False

  def enumerate_plots(self):
    # return list of plots and where/how to draw them, could be user-configured
    return [{
      'panel': y_name,
      'line': self.name,
      'x': self.names[0],
      'y': y_name,
      'exp': self,
      'width': 4 if self.is_selected else 2
    } for y_name in self.names[1:]]


class Smoother():
  def __init__(self, bandwidth, half_window=None):
    if bandwidth == 0:
      self.kernel = None
    else:
      if half_window is None:
        half_window = int(np.ceil(bandwidth * 2))
      self.kernel = np.exp(-np.arange(-half_window, half_window + 1)**2 / bandwidth**2)
    self.changed = True

  def do(self, x):
    if not isinstance(x, np.ndarray):
      x = np.array(x)
    if self.kernel is None or len(x) == 0:
      return x
    # dividing by the convolution of the kernel with a signal of all-ones handles correctly the lack of points at the edges (without biasing to a particular value)
    y = np.convolve(x, self.kernel, mode='same') / np.convolve(np.ones_like(x), self.kernel, mode='same')
    if len(self.kernel) > len(x):  # crop if larger (happens when filter is larger than signal, see np.convolve)
      start = len(y) // 2 - len(x) // 2
      y = y[start : start + len(x)]
    return y


def check_new_experiments(experiments, files_before, base_folder, window, force_reopen_files):
  # find experiment files
  files_now = set(glob.glob(base_folder + "/**/stats.csv", recursive=True))

  # add new ones
  new_files = files_now - files_before
  for filename in new_files:
    exp = Experiment(filename, base_folder, force_reopen_files)
    experiments.append(exp)
    window.add_experiment(exp, refresh_table=False)

  if len(new_files) > 0:
    window.refresh_table()

    # store the new list of files
    files_before.clear()
    files_before.update(files_now)




import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)

from itertools import product, cycle, count
from functools import partial
import heapq, logging
import numpy as np

import pyqtgraph as pg

from .plotwidget import create_plot_widget

# define lists of styles to cycle
palette = ["#4C72B0", "#DD8452", "#55A868", "#C44E52", "#8172B3", "#937860", "#DA8BC3", "#8C8C8C", "#CCB974", "#64B5CD"]
dashes = [QtCore.Qt.SolidLine, QtCore.Qt.DashLine, QtCore.Qt.DotLine, QtCore.Qt.DashDotLine, QtCore.Qt.DashDotDotLine]
dashes_by_name = dict(zip(['-', '--', ':', '-.', '-..'], dashes))
widths = [2, 1, 3]  # line widths


class Plots():
  def __init__(self, window):
    self.window = window
    window.plots = self  # back-reference

    self.panels = {}  # widgets containing plots, indexed by name (usually the plot title at the top)
    self.unused_styles = []  # reuse styles from hidden experiments. this is a heap so early styles have priority.
    self.next_style_index = 0

    pg.setConfigOptions(antialias=True, background='w', foreground='k')  # black on white

  def get_style(self):
    # reuse a previous style if possible, in order
    if len(self.unused_styles) > 0:
      return heapq.heappop(self.unused_styles)
    
    # otherwise, get a new one. start by varying color, then dashes, then line width.
    idx = self.next_style_index
    color = palette[idx % len(palette)]
    dash = dashes[(idx // len(palette)) % len(dashes)]
    width = widths[(idx // (len(palette) * len(dashes))) % len(widths)]

    self.next_style_index = idx + 1

    return (idx, {'color': color, 'style': dash, 'width': width})
  
  def drop_style(self, style_order, style):
    # return a style to the set of available ones
    heapq.heappush(self.unused_styles, (style_order, style))

  def add(self, plots):
    # e.g.: plots.add(exp.enumerate_plots())
    plotsize = None
    for plot in plots:
      # check if panel exists. there's a different panel for each x coordinate (e.g. iterations, time)
      panel_id = (plot['panel'], plot['x'])
      if panel_id not in self.panels:
        # create new panel to contain plot.
        # if title not defined, use the panel ID (e.g. stat name)
        title = str(plot.get('title', plot['panel']))
        plot_widget = create_plot_widget(title)
        panel = self.window.add_panel(plot_widget, title)

        panel.plots_dict = {}
        
        # set up mouse move event
        plot_widget.mouseMoveEvent = partial(mouse_move, panel=panel)
        plot_widget.leaveEvent = partial(mouse_leave, panel=panel)

        # mouse cursor (vertical line)
        vline = pg.InfiniteLine(angle=90, pen="#B0B0B0")
        vline.setVisible(False)
        plot_widget.getPlotItem().addItem(vline, ignoreBounds=True)  # ensure it doesn't mess autorange
        panel.cursor_vline = vline

        # mouse cursor text
        label = pg.LabelItem(justify='left')
        label.setParentItem(plot_widget.getPlotItem().getViewBox())
        label.anchor(itemPos=(0, 0), parentPos=(0, 0))
        panel.cursor_label = label

        self.panels[panel_id] = panel
      else:
        panel = self.panels[panel_id]  # reuse existing panel
      
      # get data points
      exp = plot['exp']
      (xs, ys) = (exp.data[exp.names.index(plot['x'])], exp.data[exp.names.index(plot['y'])])

      # allow overriding the style
      style = exp.style
      if 'color' in plot:
        style['color'] = plot['color']
      if 'width' in plot:
        style['width'] = plot['width']
      if 'dash' in plot and plot['dash'] in dashes_by_name:
        style['style'] = dashes_by_name[plot['dash']]
      
      try:
        pen = pg.mkPen(style)
      except:  # if the style is malformed, use the default style
        pen = pg.mkPen(exp.style)
      
      # check if plot line already exists
      if plot['line'] not in panel.plots_dict:
        # create new line
        line = panel.plot_widget.getPlotItem().plot(xs, ys, pen=pen)
        
        line.curve.setClickable(True, 8)  # size of hover region
        line.mouse_over = False
        
        panel.plots_dict[plot['line']] = line
      else:
        # update existing one
        line = panel.plots_dict[plot['line']]
        line.setData(xs, ys)
        line.setPen(pen)
  
  def remove(self, plots):
    for plot in plots:
      # find panel
      panel_id = (plot['panel'], plot['x'])
      if panel_id in self.panels:
        panel = self.panels[panel_id]

        # find plot line
        line_id = plot['line']
        if line_id in panel.plots_dict:
          # remove it
          panel.removeItem(panel.plots_dict[line_id])
          del panel.plots_dict[line_id]
        
        # if the last line was deleted, delete the panel too
        if len(panel.plots_dict) == 0:
          panel.setParent(None)
          panel.deleteLater()
          del self.panels[panel_id]

  def update_plots(self, experiments):
    # called by a timer to update plots periodically
      for exp in experiments:
        try:
          if not exp.done and next(exp.read_data):  # check if there's new data in the file
            self.add(exp.enumerate_plots())  # update plots
        except IOError:
          logging.exception('Error reading ' + exp.filename)
        except StopIteration:
          pass


def mouse_move(event, panel):
  # select curves when hovering them, and update mouse cursor
  viewbox = panel.plot_widget.getPlotItem().vb
  point = viewbox.mapSceneToView(event.pos())
  
  selected = None
  for line in panel.plots_dict.values():
    # only the first one gets selected
    inside = (not selected and line.curve.mouseShape().contains(point))

    if inside:
      selected = line
      if not line.mouse_over:
        # change line style to thicker
        line.original_pen = line.opts['pen']
        pen = pg.mkPen(line.original_pen)
        pen.setWidthF(line.original_pen.widthF() + 2)
        line.setPen(pen)

        # bring it to the front
        line.setZValue(1)

        line.mouse_over = True
    else:
      if line.mouse_over:
        # restore line style and z-order
        line.setPen(line.original_pen)
        line.setZValue(0)
        line.mouse_over = False

  # show cursor (vertical line)
  panel.cursor_vline.setVisible(True)
  x = point.x()

  if selected:
    # snap vertical line to nearest point (by x coordinate)
    data = selected.getData()
    index = np.argmin(np.abs(data[0] - x))
    (x, y) = (data[0][index], data[1][index])

    # show data coordinates
    names = [name for (name, line) in panel.plots_dict.items() if line is selected]  # ideally should return only 1
    names = ' '.join(names)
    # this trick prints floats with 3 significant digits and no sci notation (e.g. 1e-4). also consider integers.
    if x.is_integer(): x = str(int(x))
    else: x = float('%.3g' % x)
    if y.is_integer(): y = str(int(y))
    else: y = float('%.3g' % y)
    text = "%s<br/>(%s, %s)" % (names, x, y)
  else:
    text = ""

  panel.cursor_label.setText(text)  #, size='10pt'
  panel.cursor_vline.setValue(x)


def mouse_leave(event, panel):
  # hide cursor when the mouse leaves
  panel.cursor_vline.setVisible(False)





import math, logging, os, runpy
from collections import OrderedDict

import PyQt5.QtWidgets as QtWidgets
import pyqtgraph as pg

try:
  from torch import load
except ImportError:
  # fallback to regular pickle if pytorch not installed
  import pickle
  def load(path):
    with open(path, 'rb') as file:
      pickle.load(file)

try:
  from torchvision.utils import make_grid
except ImportError:
  make_grid = None

try:
  import PyQt5  # needed for matplotlib to recognize the binding to this backend
  import matplotlib
  from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
  from matplotlib.pyplot import cm as colormaps
except:
  def FigureCanvas(fig):
    logging.warning("Could not load MatPlotLib.")
    return pg.PlotWidget()
  colormaps = None
  matplotlib = None


class Visualizations():
  # custom visualizations, supports both MatPlotLib (MPL) and PyQtGraph (PG) figures
  def __init__(self, window, no_vis_snapshot, mpl_dpi):
    self.window = window
    window.visualizations = self  # back-reference

    self.panels = OrderedDict()  # widgets containing plots, indexed by name
    self.selected_exp = None  # selected experiment
    self.modules = {}  # loaded modules with visualization functions

    self.vis_counts = OrderedDict()
    self.last_size = None

    self.no_vis_snapshot = no_vis_snapshot
    if matplotlib is not None:
      matplotlib.rcParams['figure.dpi'] = mpl_dpi

  def load_single_vis(self, exp, name):
    # load a single visualization of the given experiment, with the given name.
    # first, load function info and arguments from pickled file, to the CPU.
    filename = exp.directory + '/' + name + '.pth'
    try:
      data = load(filename, map_location='cpu')
    except Exception as err:
      # ignore error about incomplete data, since file may still be written to; otherwise report it.
      if not isinstance(err, RuntimeError) or 'storage has wrong size' not in str(err):
        logging.exception("Error loading visualization data from " + filename)

    func_name = data['func']
    args = data['args']
    kwargs = data['kwargs']
    source_file = data['source']

    # call a built-in function, e.g. simple tensor visualization
    if source_file == 'builtin' and func_name == 'tensor':
      panels = tshow(*args, **kwargs, create_window=False, title=name)
    else:
      # custom visualization function. load the original source file or saved file
      # from the experiment directory, and call the specified function in it.
      if not self.no_vis_snapshot: source_file = exp.directory + '/' + name + '.py'
      panels = []

      try:
        # load module from file, unless it's cached already
        if source_file in self.modules:
          module = self.modules[source_file]
        else:
          module = runpy.run_path(source_file)

        try:
          # call the custom function, only if the module loaded successfully
          panels = module[func_name](name, *args, **kwargs)
          
          # cache module if no error so far (otherwise reload next time, maybe it's fixed)
          self.modules[source_file] = module

        except Exception:
          logging.exception('Error executing visualization function ' + func_name + ' from ' + source_file)

      except Exception:
        logging.exception('Error loading visualization function from ' + source_file)

    # ensure a list is returned
    if not isinstance(panels, list): panels = [panels]
    return panels

  def update(self):
    # called by a timer to check for updates. don't update if the experiment is finished.
    if self.selected_exp is None or self.selected_exp.done: return

    # see if the file size changed as a basic check
    try:
      new_size = os.path.getsize(self.selected_exp.directory + '/visualizations')
    except FileNotFoundError:
      return
    if self.last_size is None:
      self.last_size = new_size

    if new_size != self.last_size:
      self.last_size = new_size

      # load the list of visualizations and their refresh counts
      vis_counts = self.read_vis_counts(self.selected_exp)

      # check if any of them is new or its count changed
      for (name, count) in vis_counts.items():
        if name not in self.vis_counts or self.vis_counts[name] != count:
          # load the new data
          new_plots = self.load_single_vis(self.selected_exp, name)
          
          # assign each plot to a new or reused panel.
          # NOTE: most of this complicated logic is to deal with a specific MatPlotLib/
          # FigureCanvas bug. we cannot delete a FigureCanvas and assign a new one to the
          # same figure, or there are many graphical glitches (especially related to DPI).
          # to avoid this, we reuse the same widget (actually the parent panel widget) to
          # the same MatPlotLib figure every time (stored as overboard_panel attribute).
          new_panels = []
          old_panels = self.panels.get(name, [])
          next_idx = 0  # next available panel to reuse

          for plot in new_plots:
            if type(plot).__name__ == 'Figure' and hasattr(plot, 'overboard_panel'):
              # always reuse a previous panel for MPL figures
              panel = self.window.add_panel(plot.overboard_panel, name, reuse=True)
              panel.plot_widget.draw()  # ensure the figure is redrawn
            else:
              # not MPL, try to reuse a panel. first, skip over any existing MPL panels
              while next_idx < len(old_panels) and old_panels[next_idx].is_mpl_figure:
                next_idx += 1
              
              if next_idx < len(old_panels):
                # we can reuse this one. it contains a pg.GraphicsLayoutWidget.
                panel = old_panels[next_idx]
                panel.plot_widget.clear()
                panel.plot_widget.addItem(plot)
                panel = self.window.add_panel(panel, name, reuse=True)

              else:
                # create a new panel
                panel = self.add_vis_panel(plot, name)
            new_panels.append(panel)
          
          # remove any panels we did not reuse from the layout
          self.delete_vis_panels(list(set(old_panels) - set(new_panels)))
          self.panels[name] = new_panels

  def select(self, exp):
    # select a new experiment, showing its visualizations (and removing previously selected ones)
    self.selected_exp = exp
    self.modules = {}  # always reload modules when selecting, in case they're stale
    new_panels = []
    if exp is not None:
      # load master list of visualizations (and their counts), then load each one
      vis_counts = self.read_vis_counts(exp)

      for name in vis_counts.keys():
        # load data into plots, and turn them into visible panels
        plots = self.load_single_vis(exp, name)
        panels = [self.add_vis_panel(plot, name, add_to_layout=False) for plot in plots]
        new_panels.append((name, panels))

    new_panels = OrderedDict(new_panels)

    # remove previous widgets (this is done after loading the visualizations to reduce delay)
    self.delete_vis_panels(self.all_panels())

    # add the new widgets to the flow layout, in order
    self.panels = new_panels
    for panel in self.all_panels():
      self.window.flow_layout.addWidget(panel)

    self.last_size = None

  def all_panels(self):  # flatten nested list of panels
    if len(self.panels) == 0: return []  # special case
    return [p for panels in self.panels.values() for p in panels]
    
  def add_vis_panel(self, plot, name, add_to_layout=True):
    # wrap MatPlotLib figure or PyQtGraph PlotItem in a Qt widget
    is_mpl_figure = (type(plot).__name__ == 'Figure')
    if is_mpl_figure:
      widget = FigureCanvas(plot)
    else:
      widget = pg.GraphicsLayoutWidget()
      widget.addItem(plot)

    # wrap that in a panel with title
    panel = self.window.add_panel(widget, name, add_to_layout=add_to_layout)

    panel.is_mpl_figure = is_mpl_figure
    if is_mpl_figure:
      plot.overboard_panel = panel  # always associate the same panel with this figure

    return panel
  
  def delete_vis_panels(self, panels):
    # remove panels from the layout. MPL panels cannot be deleted, they need to
    # be reused if the same figure is displayed. note self.panels is not updated.
    for panel in panels:
      panel.setParent(None)
      if not panel.is_mpl_figure:
        panel.deleteLater()

  def read_vis_counts(self, exp):
    # read and return visualizations list, including their update counts
    vis_counts = OrderedDict()
    try:
      with open(exp.directory + '/visualizations', 'r') as file:
        for line in file:
          values = line.split('\t')
          if len(values) == 2:
            vis_counts[values[0]] = int(values[1])  # the format is: "name count\n"
    except FileNotFoundError:
      pass
    return vis_counts


tshow_images = []
def tshow(tensor, create_window=True, title='Tensor', data_range=None, grayscale=False, legend=None, **kwargs):
  """Shows a PyTorch tensor (including one or more RGB images) using PyQtGraph."""

  if make_grid is None:
    raise ImportError('Could not import torchvision (from PyTorch), which is necessary for tshow.')

  tensor = tensor.detach().cpu()
  original_shape = tensor.shape

  if data_range is None:
    data_range = (tensor.min(), tensor.max())

  if len(tensor.shape) > 4:
    logging.exception('Cannot show tensors with more than 4 dimensions.')
    return None

  # insert singleton dimensions on the left to always get 4 dimensions
  while len(tensor.shape) < 4:
    tensor.unsqueeze_(0)

  sh = tensor.shape
  if sh[0] == 1:  # case of 3D tensors, leave singleton dimension for color
    tensor = tensor.reshape(sh[1], 1, sh[2], sh[3])

  sh = tensor.shape
  if sh[1] in [1, 3]:
    # a linear collection of images: channels are RGB or single-channel.
    # choose number of columns such that they are laid out in a square.
    if 'nrow' not in kwargs:
      #kwargs['nrow'] = math.ceil(math.sqrt(sh[0]))
      kwargs['nrow'] = math.ceil(math.sqrt(sh[0]) * sh[2] / sh[3])
  else:
    # when channels are not RGB or single-channel, display them as columns
    tensor = tensor.reshape(sh[0] * sh[1], 1, sh[2], sh[3])
    kwargs['nrow'] = sh[1]


  # arrange into a grid
  image = make_grid(tensor, **kwargs, normalize=False, padding=0)
  image = image.permute(2, 1, 0).numpy()  # pytorch convention to numpy image convention

  # convert grayscale RGB images to colormapped images (single-channel)
  if tensor.shape[1] == 1:
    image = image[:,:,0]

  # show it
  im_item = pg.ImageItem(image)
  title = title + ' ' + str(tuple(original_shape))

  if create_window:  # stand-alone window
    win = pg.GraphicsWindow(title=title)
    plot = win.addPlot()
  else:  # plot item to return
    plot = pg.PlotItem()
  plot.addItem(im_item)

  plot.getViewBox().invertY(True)
  plot.setAspectLocked(True)
  plot.hideAxis('left')
  plot.hideAxis('bottom')

  # draw a grid
  (cell_w, cell_h) = (sh[3], sh[2])
  (w, h) = image.shape[0:2]
  for x in range(0, w + 1, cell_w):
    plot.plot([x, x], [0, h])
  for y in range(0, h + 1, cell_h):
    plot.plot([0, w], [y, y])

  if len(image.shape) == 3:
    # RGB image
    im_item.setLevels([data_range]*3)
  else:
    # grayscale image or heatmap, set up colormap and possibly a legend
    if not grayscale and colormaps is not None:
      # use better colormap if matplotlib is available
      colormap = colormaps.viridis
      colormap._init()
      lut = (colormap._lut * 255)[:-1,:]  # remove last row
      im_item.setLookupTable(lut)
      (low_color, high_color) = (lut[0,:3], lut[-1,:3])
    else:
      lut = []
      (low_color, high_color) = ('k', 'w')

    im_item.setLevels(data_range)

    if legend or (legend is None and not grayscale):
      # create legend with max and min values
      leg = plot.addLegend(offset=(1, 1))
      leg.addItem(FilledIcon(low_color), "Min: {:.3g}".format(data_range[0]))
      leg.addItem(FilledIcon(high_color), "Max: {:.3g}".format(data_range[1]))

      # monkey-patch paint method to draw a more opaque background
      def paint(self, p, *args):
        color = pg.mkColor(pg.getConfigOption('background'))
        color.setAlpha(200)
        p.fillRect(self.boundingRect(), pg.mkBrush(color))
      leg.paint = paint.__get__(leg)

  if create_window:
    tshow_images.append(win)  # keep reference, otherwise the window will be garbage-collected
    win.show()
  else:
    return plot


class FilledIcon(pg.graphicsItems.LegendItem.ItemSample):
  """Custom legend icon, completely filled with a single color."""
  def __init__(self, color):
    super().__init__(None)
    self.color = color
    self.setFixedWidth(20)

  def paint(self, p, *args):
    p.fillRect(self.boundingRect(), pg.mkBrush(*self.color))




#os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "TRUE"

import PyQt5.QtWidgets as QtWidgets
import PyQt5.QtCore as QtCore
import PyQt5.QtGui as QtGui
from PyQt5.QtCore import Qt
from PyQt5.Qt import QPalette, QColor

# needed right after QT imports for high-DPI screens
#QtWidgets.QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
#QtWidgets.QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)

import os

from .flowlayout import FlowLayout
from .fastslider import Slider
from .experiments import Smoother


class Window(QtWidgets.QMainWindow):
  def __init__(self, args):
    super(Window, self).__init__(parent=None)
    
    self.experiments = {}  # filled in later
    self.plots = None  # object that manages plots
    self.visualizations = None  # object that manages custom visualizations

    # get screen size
    screen_size = QtWidgets.QDesktopWidget().availableGeometry(self).size()

    # create the parent of both sidebar and plot area
    main = QtWidgets.QSplitter(self)

    # create sidebar
    sidebar = QtWidgets.QVBoxLayout()
    sidebar.setAlignment(Qt.AlignTop)
    widget = QtWidgets.QWidget(main)  # need dummy widget to wrap box layout
    widget.setLayout(sidebar)

    # plot size slider
    plotsize = args.plotsize
    if plotsize == 0:
      plotsize = screen_size.width() * 0.2
    sidebar.addWidget(QtWidgets.QLabel('Plot size'))
    slider = Slider(Qt.Horizontal)  #QtWidgets.QSlider(Qt.Horizontal)
    slider.setMinimum(screen_size.width() * 0.05)
    slider.setMaximum(screen_size.width() * 0.8)
    slider.setTickInterval(screen_size.width() * 0.005)
    slider.setValue(plotsize)  # initial value
    slider.valueChanged.connect(self.size_slider_changed)
    sidebar.addWidget(slider)
    self.size_slider = slider
    
    """# smoothness slider
    sidebar.addWidget(QtWidgets.QLabel('Smoothness'))
    slider = Slider(Qt.Horizontal)  #QtWidgets.QSlider(Qt.Horizontal)
    slider.setMinimum(0)
    slider.setMaximum(10 * 4)  # smooth_slider_changed always divides by 4 (since sliders only support ints)
    slider.setTickInterval(1)
    slider.setValue(args.smoothen)  # initial value
    slider.valueChanged.connect(self.smooth_slider_changed)
    sidebar.addWidget(slider)
    self.smooth_slider = slider
    self.smoother = Smoother(args.smoothen)"""

    # experiments list in sidebar, as a table
    table = QtWidgets.QTableWidget(0, 2)  # zero rows, two columns
    table.setHorizontalHeaderLabels(['', 'run'])  # column headers
    table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)

    # table style
    table.setShowGrid(False)
    table.setAlternatingRowColors(True)

    #table.setFocusPolicy(Qt.NoFocus)
    table.setSelectionBehavior(QtWidgets.QTableView.SelectRows)  # allow selecting rows
    table.setSelectionMode(QtWidgets.QTableView.ExtendedSelection)

    table.setHorizontalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)  # smooth scrolling
    table.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)
    
    table.verticalHeader().hide()  # hide vertical header
    
    header = table.horizontalHeader()  # configure horizontal header
    header.setStretchLastSection(True)
    header.setDefaultAlignment(Qt.AlignLeft)
    header.setSectionsMovable(True)
    header.setTextElideMode(Qt.ElideRight)  # show ellipsis for cut-off headers
    header.setHighlightSections(False)  # no bold header when cells are clicked

    table.itemClicked.connect(self.table_click)
    table.itemSelectionChanged.connect(self.table_select)

    sidebar.addWidget(table)
    self.table = table
    self.table_args = {}  # maps argument names to column indices
    #self.table_exps = {}  # maps experiment names to rows as table_exps[name].row() (even as it's sorted)
    self.prev_sort = None
    self.selected_exp = None

    # create the scroll area with plots
    (plot_scroll_widget, plot_scroll_area) = create_scroller()
    main.addWidget(plot_scroll_area)

    # finish the parent: let sidebar width remain fixed when resizing the window, while
    # the plot area can vary; and set initial sidebar size.
    main.setStretchFactor(0, 0)
    main.setStretchFactor(1, 1)
    sidebar_size = screen_size.width() * 0.15
    main.setSizes([sidebar_size, 1])

    # main layout for plots
    self.flow_layout = FlowLayout(plot_scroll_widget)
    
    self.setCentralWidget(main)

    # window size and title
    self.resize(screen_size.width() * 0.6, screen_size.height() * 0.95)
    self.setWindowTitle('OverBoard - ' + args.folder)

  def add_panel(self, widget, title, add_to_layout=True, reuse=False):
    # adds a panel to the FlowLayout (main plots display), containing a widget (e.g. FigureCanvas).
    # first, create a QGroupBox around it, to show the title
    if not reuse:
      vbox = QtWidgets.QVBoxLayout()
      vbox.addWidget(widget)
      panel = QtWidgets.QGroupBox(title)
      panel.setLayout(vbox)
      panel.plot_widget = widget  # keep a reference to the inner widget
    else:
      panel = widget  # reusing a previous panel (less common)
      panel.setTitle(title)
    
    # set the size
    plotsize = self.size_slider.value()
    panel.setFixedWidth(plotsize)
    panel.setFixedHeight(plotsize)

    if add_to_layout:  # add to window's flow layout
      self.flow_layout.addWidget(panel)

    return panel
  
  def add_experiment(self, exp, refresh_table=True):
    # ensure it has a style assigned, even if it has no plots
    (exp.style_order, exp.style) = self.plots.get_style()

    # add experiment to plots
    self.plots.add(exp.enumerate_plots())
    self.experiments[exp.name] = exp

    # add experiment to table
    (table, table_args) = (self.table, self.table_args)
    header = table.horizontalHeader()

    # disable sorting to prevent bug when adding items. will be restored by refresh_table.
    self.prev_sort = (header.sortIndicatorSection(), header.sortIndicatorOrder())  # save state
    table.setSortingEnabled(False)

    # new row
    row = table.rowCount()
    table.insertRow(row)

    # persistent mapping between an experiment and its row (even as they're sorted)
    #self.table_exps[exp.name] = QtCore.QPersistentModelIndex(table.model().index(row, 0))

    # create icon (label with unicode symbol) with the plot color. u'\u2611 \u2610'
    icon = self.set_table_cell(row, 0, u'\u2611', selectable=False)
    icon.setForeground(QColor(exp.style.get('color', '#808080')))
    # store experiment name with icon, to be retrieved later in table_click
    icon.setData(Qt.UserRole, 'hide')
    icon.setData(Qt.UserRole + 1, exp.name)

    # experiment name
    self.set_table_cell(row, 1, exp.name)

    # print a row of argument values for this experiment
    for arg_name in exp.meta.keys():
      if not arg_name.startswith('_'):
        if arg_name not in table_args:  # a new argument name, add a column
          col = table.columnCount()
          table.setColumnCount(col + 1)
          table.setHorizontalHeaderItem(col, QtWidgets.QTableWidgetItem(arg_name))  #, QtWidgets.QTableWidgetItem.Type
          table_args[arg_name] = col
        else:
          col = table_args[arg_name]
        
        self.set_table_cell(row, col, exp.meta.get(arg_name, ''))

    if refresh_table:
      self.refresh_table()
  
  def set_table_cell(self, row, col, value, selectable=True):
    # helper to set a single table cell in the sidebar.
    # try to interpret as integer or float, to allow numeric sorting of columns
    if not isinstance(value, int) and not isinstance(value, float):
      value = str(value)  # handle e.g. dicts
      try:
        value = float(value)
        if int(value) == value:  # store as int if possible, prints better
          value = int(value)
      except ValueError: pass

    if isinstance(value, str):  # faster option for strings
      item = QtWidgets.QTableWidgetItem(value)
    else:  # sort as number if not a string
      item = QtWidgets.QTableWidgetItem()
      item.setData(Qt.EditRole, QtCore.QVariant(value))

    flags = Qt.ItemIsEnabled  # set not editable
    if selectable:
      flags = flags | Qt.ItemIsSelectable
    item.setFlags(flags)
    self.table.setItem(row, col, item)
    return item

  def refresh_table(self):
    # restore sorting
    table = self.table
    table.setSortingEnabled(True)
    if self.prev_sort:  # restore sort state (which column and direction)
      table.sortItems(*self.prev_sort)

    table.resizeColumnsToContents()
    
    # resize the column headers to fully contain the header text (resizeColumnsToContents ignores the headers)
    max_width = int(0.9 * table.parentWidget().width())  # don't let any column become wider than 90% of the sidebar
    header = table.horizontalHeader()
    metrics = QtGui.QFontMetrics(table.horizontalHeaderItem(0).font())
    for col in range(table.columnCount()):
      text = table.horizontalHeaderItem(col).text()
      text_width = metrics.boundingRect(text).width()
      header.resizeSection(col, min(max_width, max(table.sizeHintForColumn(col), text_width + 20)))  # needs some extra width
  
  def sort_table_by_timestamp(self, refresh_table=True):
    # convenience function called at initialization to move the timestamp column to the start, and sort it
    if 'timestamp' in self.table_args:
      header = self.table.horizontalHeader()
      col = self.table_args['timestamp']  # (logical) column index of timestamp
      header.moveSection(col, 2)  # move to the start (after the icon and experiment name)
      self.prev_sort = (col, Qt.DescendingOrder)  # refresh_table will sort using this column and order
      
    if refresh_table:
      self.refresh_table()

  def table_click(self, item):
    # check if there's an action associated with this cell
    action = item.data(Qt.UserRole)
    if action:
      name = item.data(Qt.UserRole + 1)
      exp = self.experiments[name]
      if action == 'hide':
        # toggle visibility of a given experiment, if the icon is clicked
        if exp.visible:
          # remove all associated plots
          self.plots.remove(exp.enumerate_plots())

          # reset icon and style, allowing it to be used by other experiments
          item.setForeground(QColor(128, 128, 128))
          item.setText(u'\u2610')
          self.plots.drop_style(exp.style_order, exp.style)
          exp.style = {}
        else:
          # assign new style
          (exp.style_order, exp.style) = self.plots.get_style()
          
          item.setForeground(QColor(exp.style.get('color', "#808080")))
          item.setText(u'\u2611')

          # create plots
          self.plots.add(exp.enumerate_plots())

        exp.visible = not exp.visible

  def table_select(self):
    # selection changed
    items = self.table.selectedItems()
    if items:  # get associated experiment
      exp = self.row_to_experiment(items[0].row())
    else:
      exp = None

    # unselect previous experiment first
    if self.selected_exp and self.selected_exp != exp:
      self.selected_exp.is_selected = False
      if self.selected_exp.visible:  # update its view
        self.plots.add(self.selected_exp.enumerate_plots())

    if items:  # select new one
      exp.is_selected = True
      self.plots.add(exp.enumerate_plots())
      self.selected_exp = exp
      self.visualizations.select(exp)
    else:
      self.selected_exp = None
      self.visualizations.select(None)

  def row_to_experiment(self, row):
    # check left-most cell of the given row to get the associated experiment name
    leftmost = self.table.item(row, 0)
    name = leftmost.data(Qt.UserRole + 1)
    return self.experiments[name]
  
  def size_slider_changed(self):
    # resize plots and visualizations
    plotsize = self.size_slider.value()
    for panel in self.plots.panels.values():
      panel.setFixedWidth(plotsize)
      panel.setFixedHeight(plotsize)
    for panel_group in self.visualizations.panels.values():
      for panel in panel_group:
        panel.setFixedWidth(plotsize)
        panel.setFixedHeight(plotsize)
  
  def smooth_slider_changed(self):
    self.smoother = Smoother(self.smooth_slider.value() / 4.0)


def create_scroller():
  scroll_area = QtWidgets.QScrollArea()
  scroll_area.setWidgetResizable(True)

  scroll_widget = QtWidgets.QWidget(scroll_area)  # additional widget to enable scrollbars
  scroll_widget.setMinimumWidth(50)
  scroll_area.setWidget(scroll_widget)
  return (scroll_widget, scroll_area)


def set_style(app):
  app.setStyle("Fusion")

  directory = os.path.dirname(os.path.realpath(__file__))
  
  with open(directory + '/style.qss', 'r') as file:
    app.setStyleSheet(file.read())




import os, math, datetime, time, json, inspect, shutil

try:
  from torch import save
except ImportError:
  # fallback to regular pickle if pytorch not installed
  import pickle
  def save(obj, path):
    with open(path, 'wb') as file:
      pickle.dump(obj, file, pickle.HIGHEST_PROTOCOL)


class Logger:
  def __init__(self, directory, stat_names=None, meta=None, index_name="iteration", save_timestamp=True):
    """Initialize log writer on a new directory.
       The main file that is written is "stats.csv", containing one column for each stat."""
    if stat_names and not (all(isinstance(name, str) and not ',' in name for name in stat_names)):
      raise ValueError('stat_names must be a list of strings with no commas, if specified.')
    self.file = None
    self.directory = directory
    self.stat_names = stat_names  # can be omitted (will be set on first append() call)
    self.index_name = index_name
    self.count = 0

    # for averaging stats before appending to the log
    self.avg_accum = {}
    self.avg_count = {}

    # meta should be a dict or a Namespace object from argparse
    if meta is None:
      meta = {}  # ensure it's a new instance (default arguments all refer to the same instance)
    elif meta.__class__.__name__ == 'Namespace':  # check type without importing argparse
      meta = vars(meta)
    elif not isinstance(meta, dict):
      raise AssertionError("Meta should be a dictionary or argparse.Namespace.")
    meta['_done'] = False
    self.meta = meta
    self.save_timestamp = save_timestamp

    self.vis_functions = {}  # custom function associated with each visualization
    self.vis_counts = {}  # number of times each visualization was updated
    self.vis_padding = 0

    self.clock = -math.inf  # for rate_limit

    # create directory if it doesn't exist, and (empty) visualizations file
    os.makedirs(self.directory, exist_ok=True)
    with open(self.directory + '/visualizations', 'w') as file:
      pass

    # get current timestamp as string, including timezone offset
    if self.save_timestamp:
      self.meta['timestamp'] = str(datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0))
    
    # write arguments to JSON file
    self._save_meta()

  def append(self, points=None):
    """Write the given stats dict to CSV file. If none is given, the average values computed so far are used (see update_average)."""

    if points is None:
      # use computed average, and reset accumulator
      points = self.average()
      self.avg_accum = {}
      self.avg_count = {}

    if self.stat_names is None:  # assume the given stats are all there is
      self.stat_names = list(points.keys())

    else:  # validate them
      for name in points.keys():
        if name not in self.stat_names:
          raise ValueError('Unknown stat name: ' + name + '. Note that no new stats can be added after the first Logger.append call. Alternatively, they can be specified in the constructor.')

    if self.file is None:
      self._start_write()

    # first element is always the iteration
    self.count += 1
    self.file.write(str(self.count))

    # write remaining stat values
    for name in self.stat_names:
      self.file.write(',')
      if name in points:
        self.file.write(str(points[name]))
      else:
        self.file.write('NaN')
    self.file.write('\n')
    self.file.flush()
  
  def update_average(self, points):
    """Keep track of average value for each stat, adding a new data point."""
    for (name, value) in points.items():
      if name not in self.avg_accum:  # initialize
        self.avg_accum[name] = value
        self.avg_count[name] = 1
      else:
        self.avg_accum[name] += value
        self.avg_count[name] += 1
  
  def average(self):
    """Return the average value of each stat so far (see update_average)."""
    return {name: self.avg_accum[name] / self.avg_count[name] for name in self.avg_accum.keys()}

  def print(self, points=None, prefix=None, as_string=False, line_prefix='', line_suffix=''):
    """Print the current stats averages to the console, or the given values, nicely formatted.
    If prefix is given, only stat names beginning with it will be printed (e.g. "train" or "val")."""
    if points is None:
      points = self.average()
    if prefix:  # remove the prefix from the stat names
      points = {key[len(prefix):].strip('.'): val for (key, val) in points.items() if key.startswith(prefix)}
      line_prefix = line_prefix + prefix + ' '

    # this trick prints floats with 3 significant digits and no sci notation (e.g. 1e-4)
    text = ' '.join(["%s: %s" % (key, float('%.3g' % val)) for (key, val) in points.items()])
    text = line_prefix + text + line_suffix

    if as_string:
      return text
    else:
      print(text)
  
  def tensor(self, name, tensor, **kwargs):
    """Store a tensor to display in the OverBoard GUI. The name must be unique.
    Any keyword arguments will be passed to overboard.tshow."""
    self.visualize('tensor', name, tensor, **kwargs)

  def visualize(self, func, name, *args, **kwargs):
    """Store a visualization to display in the OverBoard GUI.
    The first argument must be a function, with the signature:
      figures = func(name, *args, **kwargs)
    where name is a unique name, and the following arguments/keyword arguments can be anything (e.g. tensors).
    The function can draw any graphics and return them as a list of MatPlotLib Figure objects or PyQtGraph PlotItem objects.
    These will be shown when the experiment is selected in the GUI."""

    if name in self.vis_functions:
      # reuse previously registered visualization function
      info = self.vis_functions[name]
      if info['func'] != func:
        raise ValueError("Attempting to register a different visualization function under a previously used name.")
      source_file = info['source']
    else:
      # it's new
      if not isinstance(name, str): raise ValueError("Visualization name must be a string.")
      if '\t' in name: raise ValueError('Visualization name cannot contain tab characters (\\t).')
      
      if func == 'tensor':
        # built-in functions, like the tensor visualization
        source_file = 'builtin'
      else:
        # user function. copy function source file to freeze changes in visualization code
        source_file = inspect.getsourcefile(func)  # python source for function
        if not source_file or func.__name__ == '<lambda>':
          raise ValueError("Only visualization functions defined at the top-level of a script are supported.")
        shutil.copy(source_file, self.directory + '/' + name + '.py')

      # register visualization function for quick look-up next time this is called
      self.vis_functions[name] = {'func': func, 'source': source_file}
      self.vis_counts[name] = 0

    # pickle the function and arguments
    func_name = func if isinstance(func, str) else func.__name__
    data = {'func': func_name, 'source': source_file, 'args': args, 'kwargs': kwargs}
    save(data, self.directory + '/' + name + '.pth')
    
    # update how many times this visualization was written, to signal a refresh in the GUI
    self.vis_counts[name] += 1
    with open(self.directory + '/visualizations', 'w') as file:
      for (key, value) in self.vis_counts.items():
        file.write('%s\t%i\n' % (key, value))
      
      # changes are detected based on file size (more reliable than file date across OS),
      # so pad with a continually changing number of spaces. when some or all visualizations
      # are updated in an iteration, the number of padding spaces will always be different.
      self.vis_padding = self.vis_padding % (len(self.vis_counts) + 1) + 1
      file.write(' ' * self.vis_padding)
  
  def rate_limit(self, seconds, reset=False):
    """Returns true once every N seconds. Can be used to limit the rate of visualizations."""
    if not reset:
      if time.time() - self.clock > seconds:
        self.clock = time.time()
        return True
      return False
    else:
      self.time = -math.inf
      return False

  # note about "with" statement/destructors:
  # this class can be used either with a "with" statement, or without.
  # the first is preferred in Python, but the second is still ok in this case for two reasons:
  # 1) circular references that prevent __del__ from being called shouldn't happen
  # 2) even if they do, append() always flushes the buffer so there's no delay, and the file
  #    will be closed anyway once the process exits.
  def __enter__(self):
    return self
  def __exit__(self, exc_type, exc_value, traceback):
    self._finish_write()
  def __del__(self):
    self._finish_write()

  def _start_write(self):
    # open CSV file and write header
    self.file = open(self.directory + '/stats.csv', 'w')
    self.file.write(self.index_name + ',' + ','.join(self.stat_names) + '\n')  # header
  
  def _finish_write(self):
    # close CSV file
    if self.file is not None:
      self.file.close()
      self.file = None

    # mark experiment as done, and write to JSON file
    if not self.meta['_done']:
      self.meta['_done'] = True
      self._save_meta()

  def _save_meta(self):
    # write metadata to JSON file
    with open(self.directory + '/meta.json', 'w') as file:
      json.dump(self.meta, file)
  